
/*
 *
 *  FASTOPEN project
 *
 *  author: Emma
 *
 *  date:  Tue Jun  7 00:58:22 CST 2011
 *
 *  revisions: none
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <net/fastopen.h>
#include <net/tcp_hack.h>

#include <crypto/rsa.h>
#include <linux/time.h>
#include <linux/crypto.h>
#include <linux/scatterlist.h>
#include <linux/gfp.h>
#include <linux/err.h>
#include <linux/ip.h>
#include <net/ip.h>

/*
 *  Module information
 */
MODULE_AUTHOR("Emma");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("FASTOPEN filters");

/*
 *  Magic number used to identify a hooked skb
 */
#define FASTOPEN_SKB_MAGIC 0x19850920

struct info{
        unsigned int sip, dip, isn;
        unsigned short sport, dport;
};

static long gettime()
{
        struct timeval t;
        do_gettimeofday(&t);
        return t.tv_sec;
}

static void pnt_array(char* name, unsigned char* array, int size)
{
        printk("%s:", name);
        int i;
        for(i = 0; i < size; i++)
                printk("%0x ", array[i] );
        printk("\n");
}

static char* inttochar(int* data_int, int size)
{
        int i, j;
        char* data_char;

        //data_char = (char *)kmalloc(sizeof(char)*size, GFP_KERNEL);
        data_char = (char *)kmalloc(sizeof(char)*size, GFP_ATOMIC);
        if(data_char == NULL){
                return 0;
        }
        memset(data_char, 0, sizeof(char)*size);
        for( i = size -1, j= 0; i >=0; i--, j++)
                data_char[i] = ((char)(data_int[j / 4] >> ((size-1-i)%4) * 8));
        return data_char;
}
char* mysha(char* c, int size)
{
        char* result;
        struct hash_desc desc;
        struct scatterlist sg[1];
        struct crypto_hash *tfm;

        tfm = crypto_alloc_hash("sha256", 0, CRYPTO_ALG_ASYNC);
        if(IS_ERR(tfm)){
                printk("err in allocating hash\n");
                return 0;
        }
        desc.tfm = tfm;

        crypto_hash_init(&desc);
        sg_init_one(sg, c, size);
        crypto_hash_update(&desc, sg, 1);

        //result = (char *)kmalloc(sizeof(char)*32, GFP_KERNEL);
        result = (char *)kmalloc(sizeof(char)*32, GFP_ATOMIC);
        if(result == NULL){
                crypto_free_hash(tfm);
                return 0;
        }
        memset(result, 0, sizeof(char)*32);

        crypto_hash_final(&desc, result);
        crypto_free_hash(tfm);
        return result;
}
static rsa_op* mkrc(struct info* rc, u8* sp_data, u32 data_len)
{
        rsa_op *nc, *dc, *c, *m;//, *ec, *m2;
        int i;
        long t = gettime();
//printk("in mkrc: sip %0x, dip %0x, sport %0x, dport %0x, isn %0x, t %0x\n", 
  //    rc->sip, rc->dip, rc->sport, rc->dport, rc->isn, t);

        int ec_int = 0x10001;
        char *data_ec = (char *)&ec_int;

        char data_nc[] = {0xb8,0xce,0x05,0xdb,0x5f,0xee,0xb2,0x40,0x9d,0xe1,0x0a,0xb1,0xff,0x63,0x9c,0x32,0x0c,0x91,0x71,0x9f,0x8d,0x9d,0x3a,0x76,0xcd,0x78,0xdf,0x88,0xef,0x99,0x6f,0xcb,0x65,0xfe,0x2c,0x38,0x39,0x69,0xae,0x5f,0x61,0x21,0x45,0xec,0x75,0x46,0x30,0x1f,0xa2,0x55,0xc9,0x9b,0x0e,0xc1,0x28,0x65,0x4e,0x3e,0xa2,0x47,0x50,0xea,0x49,0x87};
        char data_dc[] = {0x8f,0xa4,0xb3,0xb0,0xbc,0x74,0x50,0x61,0x21,0x56,0x56,0x75,0xdc,0x48,0x94,0x93,0xe2,0x68,0xfa,0xa4,0x45,0xf2,0xbc,0x0a,0xb3,0xd3,0x0c,0x0c,0xe8,0x0f,0x9f,0x70,0x6b,0x3a,0xbe,0x91,0x14,0x43,0x8e,0x63,0xb2,0x67,0x1b,0x49,0xa4,0x8b,0xa8,0xb9,0x4c,0x68,0xc7,0x46,0x17,0x54,0xe3,0x95,0xcb,0x19,0x6d,0x8a,0x17,0xb9,0x87,0x01};
       unsigned char data[16 + data_len];
        char *data_sport = (char *)&rc->sport;
        for (i = 0; i < 2; i++)
                data[i] = *(data_sport + i);
        char *data_dport = (char *)&rc->dport;
        for (i = 0; i < 2; i++)
                data[2 + i] = *(data_dport + i);
        char *data_sip = (char *)&rc->sip;
        for (i = 0; i < 4; i++)
                data[4 + i] = *(data_sip + i);
        char *data_dip = (char *)&rc->dip;
        for (i = 0; i < 4; i++)
                data[8 + i] = *(data_dip + i);
        char *data_isn = (char *)&rc->isn;
        for (i = 0; i < 4; i++)
                data[12 + i] = *(data_isn + i);
        if(sp_data)
                for (i = 0; i < data_len; i++)
                        data[16 + i] = *(sp_data + i);

        //pnt_array("ori_rc_data", data, 16+data_len);
        char* hash = mysha(data, 16+data_len);
        //pnt_array("ori_hash", hash, 32);
        unsigned char data_m[40];

        for (i = 0; i < 4; i++)
                data_m[i] = 0;
        for (i = 0; i < 32; i++)
                data_m[i + 4] = hash[i];
        char *data_t = (char *)&t;
        for (i = 0; i < 4; i++)
                data_m[i + 36] = *(data_t + i);
        //pnt_array("data in mkrc", data_m, 40);

        kfree(hash);

        rsa_op_init(&nc, data_nc, 64, 0);
        rsa_op_init(&dc, data_dc, 64, 0);
        rsa_op_init(&m, data_m, 64, 0);
        //printk("m:");
        //rsa_op_print(m, 1);

        rsa_op_alloc(&c, 16);
printk("before cipher rc\n");
        rsa_cipher(&c, m, dc, nc);
printk("after cipher rc\n");
        rsa_op_free(nc);
        rsa_op_free(dc);
        rsa_op_free(m);
        return c;
}

char* msgc(struct info* rc_info, u8* sp_data, u32 data_len)
{
        rsa_op* rc;
printk("in msgc\n");
        char data_pc[128] = {0x7d,0x7b,0x99,0xd3,0xd7,0x9b,0xc7,0x5d,0x75,0x2d,0xfd,0xb7,0x3e,0x2a,0x7d,0xb9,0xb,0xe7,0xdd,0xa2,0x84,0xcd,0xde,0xc9,0xc5,0xf7,0x2a,0x11,0xf1,0xf2,0xfe,0xd1,0x76,0xf,0xa4,0x2e,0xb1,0x6a,0x25,0xda,0xbb,0x6d,0xf2,0x78,0xd5,0x8d,0xb5,0x54,0x84,0x68,0x35,0x48,0x61,0xbc,0x86,0xbb,0x7c,0x56,0x1,0x27,0x94,0x12,0xc9,0xec,0x7a,0xd,0x83,0x43,0xd7,0xdd,0x5c,0x7c,0x5,0xf5,0x8f,0xa8,0xb8,0x1e,0xc7,0x30,0x24,0xb8,0xd1,0xbf,0x2e,0x5b,0xe3,0x39,0x55,0x2f,0x9c,0xe3,0xf7,0x3,0xb,0xdd,0xde,0xfd,0x9,0xa8,0xe0,0x6f,0x55,0x29,0x48,0xbb,0xdc,0x4,0x57,0x77,0xf3,0xe6,0x34,0xbc,0xea,0xd9,0x19,0xb2,0xba,0x5d,0xc7,0x89,0xba,0x35,0x41,0xdb,0xd7,0x31};
        rc = mkrc(rc_info, sp_data, data_len);
//      rsa_op_print(rc, 1);
        char* data_rc = inttochar(rc->data, 64);
//      pnt_array("data_rc", data_rc, 64);
        char* msgc = (char *)kmalloc(sizeof(char)*192, GFP_ATOMIC);
        if(msgc == NULL){
                return 0;
        }
//        memset(msgc, 0, sizeof(char)*32);

        memcpy(msgc, data_pc, 128);
        memcpy(msgc + 128, data_rc, 64);

     //   kfree(data_pc);
        kfree(data_rc);

        rsa_op_free(rc);

  //      pnt_array("msgc", msgc, 192);
        return msgc;
}

static char* decpc(char* pc, int ip, long t)
{
        rsa_op *np, *ep, *c, *m;
        int i, ts, sip;
        unsigned char dur;

        int ep_int = 0x10001;
        char *data_ep = (char *)&ep_int;
        char data_np[] = {0xb7,0x70,0x45,0xd7,0xcc,0x2f,0x5a,0xb5,0x24,0x5f,0xf2,0xa7,0x37,0x8d,0x65,0xe0,0x2d,0x2c,0x43,0x6b,0xed,0x2a,0x4c,0xae,0xce,0xb6,0x19,0xde,0x10,0xeb,0xd3,0xce,0xc0,0x6b,0x1c,0xdb,0x0d,0xee,0xae,0xc0,0xe8,0x2b,0xa0,0xdd,0x49,0xa1,0x09,0x61,0x3b,0xe8,0xa3,0xe2,0xce,0x78,0x1c,0x5b,0xac,0x4f,0xde,0x69,0xff,0x27,0xac,0xec,0xdd,0xf0,0x8c,0x0b,0x87,0xce,0x93,0xb4,0x46,0xd0,0x96,0xa4,0x32,0x42,0x3d,0x9c,0x73,0x35,0x2d,0xb9,0xf4,0x35,0x94,0x8c,0x90,0x2d,0x3d,0x57,0x0b,0xa5,0x57,0xcd,0x97,0x6f,0x7b,0x82,0x72,0xa3,0xef,0x19,0xcd,0x73,0x26,0xb6,0x48,0x3c,0xc2,0x3f,0x97,0x9a,0x2d,0x4b,0xbe,0x56,0xb1,0xa4,0xc1,0x44,0xc5,0x65,0x9b,0x0f,0xe6,0xc1};
        unsigned char data_c[128];
        for (i = 0; i < 128; i++)
                data_c[i] = *(pc + i);
        //pnt_array("pc", data_c, 128);

        rsa_op_init(&np, data_np, 128, 0);
        rsa_op_init(&ep, data_ep, 3, 0);
        rsa_op_init(&c, data_c, 128, 0);

        //printk("ep:");
        //rsa_op_print(ep, 1);
        //printk("np:");
        //rsa_op_print(np, 1);
//        printk("c:");
//        rsa_op_print(c, 1);
        rsa_op_alloc(&m, 32);
printk("before cipher pc\n");
        rsa_cipher(&m, c, ep, np);
printk("after cipher pc\n");
        unsigned char* data_m = (unsigned char *)kmalloc(sizeof(char)*128, GFP_ATOMIC);
        if(data_m == NULL){
                return 0;
        }
        memset(data_m, 0, sizeof(char)*128);
        data_m = inttochar(m->data, 128);
        //pnt_array("mdata", data_m, 128);
        sip = *(int *)(data_m + 68);
        ts = *(int *)(data_m + 72);
        dur = *(data_m + 76);
printk("t %0x, ts %0x, dur %0x, sip %0x, ip %0x\n", t, ts, dur, sip, ip);
        //if (t < ts || t > ts + dur * 3600){
        if (t < ts){
                kfree(data_m);
                return NULL;
        }
        rsa_op_free(np);
        rsa_op_free(ep);
        rsa_op_free(c);
        rsa_op_free(m);

        return data_m;
}
static int decrc(char* rc, struct info* rc_info, char* cpub, long t, char* sp_data, u32 data_len)
{
        rsa_op *nc, *ec, *c, *m;
        int i;
        long ts;
//      pnt_array("cpub", cpub, 68);
        //pnt_array("rc", rc, 64);
        rsa_op_init(&ec, cpub, 3, 0);
        rsa_op_init(&nc, (cpub + 4), 64, 0);
  //      printk("nc:");
  //      rsa_op_print(nc, 1);
  //     printk("ec:");
  //      rsa_op_print(ec, 1);

printk("in decrc: sip %0x, dip %0x, sport %0x, dport %0x, isn %0x, t %0x\n",
      rc_info->sip, rc_info->dip, rc_info->sport, rc_info->dport, rc_info->isn, t);
        rsa_op_alloc(&m, 16);

        rsa_op_init(&c, rc, 64, 0);

printk("before cipher rc\n");
        rsa_cipher(&m, c, ec, nc);
printk("after cipher rc\n");
//printk("dec_m:");
  //     rsa_op_print(m, 1);

        unsigned char* data_m = inttochar(m->data, 128);
        ts = *(int *)(data_m + 100);
        printk("t %0x, ts %0x\n", t, ts);

        //if (t < ts || t > ts + 30000){
        if (t < ts - 300){
                kfree(cpub);
                kfree(data_m);
                return -1;
        }
        char* hash_rc = data_m + 68;
        unsigned char data[16+data_len];
        char *data_sport = (char *)&rc_info->sport;
        for (i = 0; i < 2; i++)
                data[i] = *(data_sport + i);
        char *data_dport = (char *)&rc_info->dport;
        for (i = 0; i < 2; i++)
                data[2 + i] = *(data_dport + i);
        char *data_sip = (char *)&rc_info->sip;
        for (i = 0; i < 4; i++)
                data[4 + i] = *(data_sip + i);
        char *data_dip = (char *)&rc_info->dip;
        for (i = 0; i < 4; i++)
                data[8 + i] = *(data_dip + i);
        char *data_isn = (char *)&rc_info->isn;
        for (i = 0; i < 4; i++)
                data[12 + i] = *(data_isn + i);
        for (i = 0; i < data_len; i++)
                data[16 + i] = *(sp_data + i);


      //pnt_array("data", data, 16 + data_len);
        char* hash = mysha(data, 16+data_len);
      //pnt_array("hash", hash, 32);
      //pnt_array("hash_rc", hash_rc, 32);
        if(memcmp(hash, hash_rc, 32)){
                kfree(hash);
                kfree(data_m);
                kfree(cpub);
                return 2;
        }

        kfree(hash);
        kfree(cpub);
        kfree(data_m);
        return 0;
}

int dec(struct info* rc_info, char* pc_rc, char* data, u32 data_len)
{
        long t = gettime();
printk("before decryption\n");
//printk("in dec sport: %0x, dport: %0x, sip %0x, dip %0x, isn %0x\n", rc_info->sport, rc_info->dport, rc_info->sip, rc_info->dip, rc_info->isn);
        char* cpub = decpc(pc_rc, rc_info->sip, t);
               //      pnt_array("cpub", cpub, 68);
        int ret;
        if (!cpub) return -1;
        else
//              ret = decrc(pc_rc + 128, rc_info, cpub, t);
                if ((ret = decrc(pc_rc + 128, rc_info, cpub, t, data, data_len)) != 0)
//                      printk("decrc err\n");
                                return -1;

printk("after decryption\n");
        return 0;
}

static int skb_is_fastopen(struct tcphdr *th)
{
	int length = th->doff * 4 - sizeof(struct tcphdr);
	unsigned char *ptr = (unsigned char *)(th + 1);

	while (length > 0) {
		int opcode = *ptr++;
		int opsize;

		switch (opcode) {
		case TCPOPT_EOL:
			return 0;
		case TCPOPT_NOP:
			length--;
			continue;
		case TCPOPT_FASTOPEN:
			return 1;
		default:
			opsize = *ptr++;
		}
		ptr += opsize - 2;
		length -= opsize;
	}
	return 0;
}

static void dump_ip_header(struct iphdr* iph)
{
	printk("ihl: %u, version: %u\n", iph->ihl, iph->version);
	printk("tos: %u, totlen: %u\n", iph->tos, ntohs(iph->tot_len));
	printk("id: %u, frag_off: %u\n", ntohs(iph->id), ntohs(iph->frag_off));
	printk("ttl: %u, protocol: %u\n", iph->ttl, iph->protocol);
	printk("check: %u, saddr: %x, daddr: %x\n", ntohs(iph->check),
			iph->saddr, iph->daddr);
}

static void dump_tcp_header(struct tcphdr* th)
{
	printk("src port: %u, dst port: %u\n", ntohs(th->source), ntohs(th->dest));
	printk("seq: %u, ack_seq: %u\n", ntohl(th->seq), ntohl(th->ack_seq));
	printk("doff: %u\n", th->doff);
	printk("syn: %u\n", th->syn);
	printk("ack: %u\n", th->ack);
	printk("rst: %u\n", th->rst);
	printk("fin: %u\n", th->fin);
	printk("psh: %u\n", th->psh);
	printk("urg: %u\n", th->urg);
	printk("ece: %u\n", th->ece);
	printk("cwr: %u\n", th->cwr);
	printk("window %u, check %x, urg %u\n", ntohs(th->window),
			ntohs(th->check), ntohs(th->urg_ptr));
	{
		int i;
		printk("options at:\n");
		for (i = 5; i < th->doff + 1; i++) {
			printk("%x ", *((u32*)(th) + i));
		}
		printk("\n");
	}
}

static __sum16 tcp_v4_checksum_init(struct sk_buff *skb)
{
	const struct iphdr *iph = ip_hdr(skb);

	if (skb->ip_summed == CHECKSUM_COMPLETE) {
		if (!csum_tcpudp_magic(iph->saddr, iph->daddr,
				skb->len, IPPROTO_TCP, skb->csum)) {
			skb->ip_summed = CHECKSUM_UNNECESSARY;
			return 0;
		}
	}

	skb->csum = csum_tcpudp_nofold(iph->saddr, iph->daddr,
			skb->len, IPPROTO_TCP, 0);

	if (skb->len <= 76) {
		return __skb_checksum_complete(skb);
	}
	printk("this path\n");
	return 0;
}

static void __tcp_v4_send_check(struct sk_buff *skb, __be32 saddr, __be32 daddr)
{
	struct tcphdr *th = tcp_hdr(skb);
	if (skb->ip_summed == CHECKSUM_PARTIAL) {
		th->check = ~tcp_v4_check(skb->len, saddr, daddr, 0);
		skb->csum_start = skb_transport_header(skb) - skb->head;
		skb->csum_offset = offsetof(struct tcphdr, check);
	} else {
		th->check = tcp_v4_check(skb->len, saddr, daddr,
					 csum_partial(th,
						      th->doff << 2,
						      skb->csum));
	}
}

/*
 *      Copy the special data from the socket control block
 *
 *      Question: are we thread-safe here with respect to:
 *              1. setsockopt behavior
 *              2. clean up of the special data due to connection
 *              established
 */
static int moving_special_data(struct sk_buff *skb, u32 enc_len)
{
	u8* sp_data = skb->special_data;
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th = tcp_hdr(skb);

  	struct info rc_info;

  	rc_info.sip = iph->saddr;
   	rc_info.dip = iph->daddr;
   	rc_info.sport = th->source;
   	rc_info.dport = th->dest;
   	rc_info.isn = th->seq;

        if (unlikely(skb_tailroom(skb) < skb->special_data_len + 196)) {
		printk("special len: %d\n", skb->special_data_len);
		skb->special_data_len = 0;
		skb->special_data = NULL;
		return -1;
	}
	/*
	 *	Copy the special data to the socket.
	 *	We don't care about the correctness of the data copied,
	 *	user should not call connect at the same time with
	 *	another thread calling the setsockopt.
	 */

		__be32 *magic_ptr = (__be32 *) skb->tail;
		u8* cur_tail	= (u8*) (magic_ptr + 3);
		printk("skb tail = %p, len = %u\n", skb->tail, skb->len);
		printk("th end = %p\n", ((char*)th) + th->doff * 4);
		dump_tcp_header(th);
		/*
		 *	Don't consider the memory alignment firstly
		 */
		*magic_ptr++	= htonl(FASTOPEN_SKB_MAGIC);
		*magic_ptr++	= htonl(skb->special_data_len);
		*magic_ptr++	= htonl(enc_len);

		/*
		 *
		 *
		 *
		 *
		 *	@EE:
		 *
		 *		here sp_data contains the special data pointer,
		 *		enc it. Remember to judge the NULL case
		 *
		 *	
		 *	YOU NEED COMPUTE enc_data and new special data here
		 *
		 *
		 */
		char* enc_data = msgc(&rc_info, sp_data, skb->special_data_len);
		
		if (sp_data)
			memcpy(cur_tail, sp_data, skb->special_data_len);
		if (enc_len)
			memcpy(cur_tail + skb->special_data_len, enc_data, enc_len);
		__skb_put(skb, skb->special_data_len + sizeof(__be32) * 3 + enc_len);
		skb->truesize += skb->special_data_len + sizeof(__be32) * 3 + enc_len;
		iph->tot_len = htons(skb->len);
		/*
		 *	@BUG FIX:
		 *		remember to clear th->check before recheck!
		 */
		th->check = 0;
		__tcp_v4_send_check(skb, iph->saddr, iph->daddr);
		ip_send_check(iph);
		skb->special_data = NULL;
		skb->special_data_len  = 0;
		dump_ip_header(iph);
		dump_tcp_header(th);
	return 0;
}

/*
 *	Fetching the special data into the skb, and extracting the 
 *	encrypting data
 */
static int fetching_special_data(struct sk_buff *skb, struct tcphdr *th,
				u8 **enc_data_ptr, u32 *enc_len_ptr)
{
	struct info rc_info;
   	struct iphdr* iph = ip_hdr(skb);
   	rc_info.sip = iph->saddr;
   	rc_info.dip = iph->daddr;
   	rc_info.sport = th->source;
   	rc_info.dport = th->dest;
   	rc_info.isn = th->seq;


	__be32 *magic_hdr = (__be32 *) ((char *)th + th->doff * 4);
	if (unlikely(ntohl(*magic_hdr) != FASTOPEN_SKB_MAGIC))
		return -1;
	{
	//	struct iphdr *iph = ip_hdr(skb);
		u32 sp_data_len	= ntohl(*(magic_hdr + 1));
		u32 enc_data_len = ntohl(*(magic_hdr + 2));
		u32 datalen;
		u8 *sp_data;

		if (sp_data_len > FASTOPEN_MAX_CONN_DATA) {
			printk("too large a conn data segment: %u\n",
							sp_data_len);
			return -1;
		}
		skb->special_data = NULL;
		skb->special_data_len = 0;

		if (sp_data_len) {
			sp_data = kmalloc(sp_data_len, GFP_ATOMIC);
			printk("allocate a pointer at %p %u\n", sp_data, sp_data_len);
			if (!sp_data) {
				printk("no memory\n");
				return -1;
			}
			memcpy(sp_data, magic_hdr + 3, sp_data_len);
		} else {
			sp_data = NULL;
		}

		*enc_data_ptr = (u8*) (magic_hdr + 3) + sp_data_len;
		*enc_len_ptr = enc_data_len;



		/*
		 *
		 *
		 *
		 *
		 *	@EE:
		 *	Now *enc_data_ptr points to the enc data
		 *	    *enc_len_ptr points to its length (192?)
		 *
		 *	special data is contained in sp_data, you can use it directly
		 *
		 *
		 *
		 *
		 */

        	if (dec(&rc_info, *enc_data_ptr, sp_data, sp_data_len))
                	return -1;


		skb->special_data = sp_data;
		skb->special_data_len = sp_data_len;

		/*
		 *	The encrpyting data is still valid since we
		 *	just update len field :-)
		 */
		datalen = skb->tail - (u8 *)magic_hdr;
		skb->truesize -= datalen;
		skb->len -= datalen;
		iph->tot_len = htons(skb->len);
		skb->tail = (u8 *)magic_hdr;
	//	kfree(sp_data);
	}
	return 0;
}


/*
 *  This is the netfilter hook processing the output segments
 *
 *
 *  @Current function:
 *    check the tcp header, if it tags with fastopen and SYN, then
 *    add a "hello world\n" to the data area of skb and transmit it
 */
static unsigned int fastopen_encrypt_out(unsigned int hooknum,
				     struct sk_buff *skb,
				     const struct net_device *in,
				     const struct net_device *out,
				     int (*okfn) (struct sk_buff *))
{
	/*
	 *  The output data should have transport header
	 */
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th = tcp_hdr(skb);

	if (iph->protocol != IPPROTO_TCP)
		goto out;

	if (th == NULL) {
		printk("why th should be NULL?\n");
		goto out;
	}

	if (th->syn) {
		printk("received a syn packet\n");
		if (!skb_is_fastopen(th))
			goto out;

		if (moving_special_data(skb, 192) < 0) {
			printk("fail to assign the special data\n");
			goto out;
		}
		/*
		 *  Wrapper use sk to get source and dest address, we first build
		 *  tcp header checksum, then build ip header checksum
		 */
		printk("check %x %x %x\n", th->check, iph->check, skb->csum);

	}

 out:
	return NF_ACCEPT;

}

/*
 *  This is the netfilter hook processing the input segments
 *
 *  @Current function:
 *    check the tcp header, if it tags with fastopen and SYN, then
 *    finds out the data it carries and outputs that
 */
static unsigned int fastopen_decrypt_in(unsigned int hooknum,
				    struct sk_buff *skb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn) (struct sk_buff *))
{
	struct iphdr *iph = ip_hdr(skb);
	/* we don't have transport header data set currently */
	struct tcphdr *th = (struct tcphdr *)((char *)iph + iph->ihl * 4);

	if (iph->protocol != IPPROTO_TCP)
		goto out;

	if (th == NULL) {
		printk("why th should be NULL?\n");
		goto out;
	}

	if (th->syn) {
		u8 *enc_data = NULL;
		u32 enc_data_len;

		if (!skb_is_fastopen(th)) {
			goto out;
		}

		/*
		 *  We should first check if the check sum is correct
		 */
		dump_ip_header(iph);
		dump_tcp_header(th);
		printk("skb size %d, truesize %d\n", skb->len, skb->truesize);
		printk("skb head %p, data %p, tail %p\n", skb->head, skb->data, skb->tail);

		if (!skb_csum_unnecessary(skb) && tcp_v4_checksum_init(skb))
			goto drop;

		if (!skb_csum_unnecessary(skb) && __skb_checksum_complete(skb))
			goto drop;

		skb->ip_summed = CHECKSUM_UNNECESSARY;
		if (fetching_special_data(skb, th, &enc_data, &enc_data_len) < 0) {
			printk("error in getting special data\n");
			goto drop;
		}
		printk("fastopen syn accepted: special data len: %d\n",
				skb->special_data_len);
	}
 out:
	return NF_ACCEPT;
 drop:
	return NF_DROP;
}

static struct nf_hook_ops fastopen_client_hook = {
	.hook = fastopen_encrypt_out,
	.pf = PF_INET,
	.hooknum = NF_INET_LOCAL_OUT,
	.priority = NF_IP_PRI_FIRST,
	.owner = THIS_MODULE
};

static struct nf_hook_ops fastopen_server_hook = {
	.hook = fastopen_decrypt_in,
	.pf = PF_INET,
	.hooknum = NF_INET_LOCAL_IN,
	.priority = NF_IP_PRI_FIRST,
	.owner = THIS_MODULE
};

static int __init init_fastopen(void)
{
	nf_register_hook(&fastopen_client_hook);
	nf_register_hook(&fastopen_server_hook);
	printk("hello emma!\n");
	return 0;
}

static void __exit clean_fastopen(void)
{
	nf_unregister_hook(&fastopen_client_hook);
	nf_unregister_hook(&fastopen_server_hook);
	printk("goodbye emma!\n");
}

module_init(init_fastopen);
module_exit(clean_fastopen);
