
#include "fastopen_kernel_def.h"

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "tst.h"
#include <pthread.h>

/*
 *	Thread version for the FASTOPEN connectivity test
 */


static char *msg_pref_client = "client-pref";
static char *msg_pref_server = "server-repl";

static int work_client(int fd)
{
	char buf[100];
	printf("her %d\n", fd);
	if (stable_recv(fd, buf, strlen(msg_pref_server) + 1) >= 0) {
		assert(strcmp(msg_pref_server, buf) == 0);
		int i = 0;
		for (i = 0; i < 1000; i++) {
			char buf_ans[100];
			sprintf(buf, "client-tst-%d", i);
			if (stable_send(fd, buf, strlen(buf) + 1) < 0)
				break;
			sprintf(buf_ans, "server-tst-%d", i);
			if (stable_recv(fd, buf, strlen(buf_ans) + 1) < 0)
				break;
			assert(strcmp(buf_ans, buf) == 0);
		}
	} else {
		printf("receive error");
	}
	close(fd);
	return 0;
}

static int work_server(int serv)
{
	int fastopen_type = 1;
	assert(setsockopt(serv, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_type, sizeof(int))
	       == 0);
	int len = -1;
	unsigned char data[1024];
	socklen_t lenlen = sizeof(len);
	assert(getsockopt
	       (serv, IPPROTO_TCP, TCP_FASTOPEN_CONN_DATA_LEN, &len, &lenlen) == 0);
	struct fastopen_conn_data_req req = {
		.buf = data,
		.len = len
	};
	socklen_t reqlen = sizeof(req);
	printf("request length = %d\n", req.len);
	assert(getsockopt(serv, IPPROTO_TCP, TCP_FASTOPEN_CONN_DATA, &req, &reqlen)
	       == 0);
	printf("received %d, %s\n", req.len, req.buf);
	assert(strcmp(msg_pref_client, (char*) req.buf) == 0);

	if (stable_send(serv, msg_pref_server, strlen(msg_pref_server) + 1) >=
	    0) {
		int i = 0;
		for (i = 0; i < 1000; i++) {
			char buf[100];
			char buf_ans[100];
			sprintf(buf_ans, "client-tst-%d", i);
			if (stable_recv(serv, buf, strlen(buf_ans) + 1) < 0)
				break;
			assert(strcmp(buf_ans, buf) == 0);
			sprintf(buf, "server-tst-%d", i);
			if (stable_send(serv, buf, strlen(buf) + 1) < 0)
				break;
		}
		printf("test passed\n");
	}
	close(serv);
	return 0;
}

static void *sthread(void *targ)
{
	int fd = (int)targ;
	return (void *)work_server(fd);
}

static int server_thread(int fd)
{
	pthread_t pid;
	if (pthread_create(&pid, NULL, sthread, (void*) fd) < 0) {
		printf("server: can not create thread for %d\n", fd);
		return -1;
	}
	if (pthread_detach(pid) < 0) {
		printf("server: error in detaching the thread for %d\n", fd);
		return -1;
	}
	return 0;
}

static void* cthread(void* targ)
{
	int fd = (int) targ;
	printf("fd is %d\n", fd);
	return (void*) work_client(fd);
}

static int client_thread(int fd)
{
	pthread_t pid;
	if (pthread_create(&pid, NULL, cthread, (void*) fd) < 0) {
		printf("client: can not create thread for %d\n", fd);
		return -1;
	}
	if (pthread_detach(pid) < 0) {
		printf("client: error in detaching the thread for %d\n", fd);
		return -1;
	}
	return 0;
}


int main(int argc, char **argv)
{
	if (argc > 1 && argv[1][0] == 'c') {
		int client_count = 1000;
		if (argc > 2) {
			client_count = atoi(argv[2]);
		}
		while (client_count-- > 0) {
			int cli = create_fastopen_client("192.168.56.101", msg_pref_client);
			client_thread(cli);
		}
		sleep (20);
	} else {
		int ser = create_fastopen_server();
		server_loop(ser, server_thread);
	}
	return 0;
}
