
#ifndef _FASTOPEN_KERNEL_DEF_H_
#define _FASTOPEN_KERNEL_DEF_H_


#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)

typedef unsigned char u8;

#define TCP_HACKING		19	/* @EE */
#define TCP_FASTOPEN		20	/* @EE */
#define TCP_FASTOPEN_CONN_ADDR      21
#define TCP_FASTOPEN_CONN_DATA      22
#define TCP_FASTOPEN_CONN_DATA_LEN  23

#define FASTOPEN_MAX_CONN_DATA      1024
struct fastopen_conn_data_req {
        int     len;
        u8      *buf;
};

/*
 *  This function compare content in *val_ptr to comp_val
 *  if equal, set *val_ptr to set_val, and return comp_val
 *  else do nothing but return *val_ptr
 *
 *  above thing is ensured to be in one atomic operation
 */
static inline int atomic_cmpxchg(int* val_ptr, int comp_val, int set_val)
{
  return __sync_val_compare_and_swap(val_ptr, comp_val, set_val);
}


#if defined(__X86__)
#include <asm/unistd_32.h>
#elif defined(__X86_64__)
#include <asm/unistd_64.h>
#else
#error "don't support such architecture\n"
#endif


#include <linux/net.h> /* for socket call number */
#include <errno.h>

/*
 *	Assembly entry to the call
 */
extern int syscall0(int n);
extern int syscall1(int n, int);
extern int syscall2(int n, int, int);
extern int syscall3(int n, int, int, int);
extern int syscall4(int n, int, int, int, int);


#define err_syscall(fn) ({	\
	int ret = (fn);		\
	errno = -ret;		\
	ret < 0? -1: ret;	\
	})


#define real_socket(a1, a2, a3)	({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_SOCKET, args));		\
	})

#define real_connect(a1, a2, a3)	({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_CONNECT, args));		\
	})

#define real_send(a1, a2, a3, a4)	({			\
	void* args[] = {a1, a2, a3, a4};				\
	err_syscall(syscall2(__NR_socketcall, SYS_SEND, args));		\
	})

#define real_sendmsg(a1, a2, a3)	({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_SENDMSG, args));		\
	})

#define real_recv(a1, a2, a3, a4)	({			\
	void* args[] = {a1, a2, a3, a4};				\
	err_syscall(syscall2(__NR_socketcall, SYS_RECV, args));		\
	})

#define real_recvmsg(a1, a2, a3)	({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_RECVMSG, args));		\
	})

#define real_accept(a1, a2, a3)	({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_ACCEPT, args));		\
	})

#define real_accept4(a1, a2, a3, a4)	({			\
	void* args[] = {a1, a2, a3, a4};				\
	err_syscall(syscall2(__NR_socketcall, SYS_ACCEPT4, args));		\
	})

#define real_getsockopt(a1, a2, a3, a4, a5) ({		\
	void* args[] = {a1, a2, a3, a4, a5};			\
	err_syscall(syscall2(__NR_socketcall, SYS_GETSOCKOPT, args));	\
	})

#define real_setsockopt(a1, a2, a3, a4, a5)	 ({		\
	void* args[] = {a1, a2, a3, a4, a5};			\
	err_syscall(syscall2(__NR_socketcall, SYS_SETSOCKOPT, args));	\
	})

#define real_getpeername(a1, a2, a3) ({			\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_GETPEERNAME, args));	\
	})

#define real_bind(a1, a2, a3) ({				\
	void* args[] = {a1, a2, a3};				\
	err_syscall(syscall2(__NR_socketcall, SYS_BIND, args));		\
	})
	

#define real_write(a1, a2, a3)		err_syscall(syscall3(__NR_write, a1, a2, a3))
#define real_writev(a1, a2, a3)		err_syscall(syscall3(__NR_writev, a1, a2, a3))
#define real_read(a1, a2, a3)		err_syscall(syscall3(__NR_read, a1, a2, a3))
#define real_readv(a1, a2, a3)		err_syscall(syscall3(__NR_readv, a1, a2, a3))
#define real_close(a1)			err_syscall(syscall1(__NR_close, a1))


#endif
