
#include "fastopen_kernel_def.h"

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "tst.h"
#include <pthread.h>

/*
 *	IO multiplexing test
 */

static char *msg_pref_client = "client-pref";
static char *msg_pref_server = "server-repl";

static int work_client(int fd)
{
	char buf[100];
	printf("her %d\n", fd);
	strcpy(buf, msg_pref_client);
	assert(stable_send(fd, buf, strlen(msg_pref_client) + 1) ==
			strlen(msg_pref_client) + 1);
	assert(stable_recv(fd, buf, strlen(msg_pref_server) + 1) ==
			strlen(msg_pref_server) + 1);
	assert(strcmp(buf, msg_pref_server) == 0);
	close(fd);
	return 0;
}

static int work_server(int serv)
{
	char data[1024];

	assert(test_select(serv) == 1);
	assert(test_poll(serv) == 1);
	assert(test_epoll(serv) == 1);

	assert(stable_recv(serv, data, strlen(msg_pref_client) + 1)
			== strlen(msg_pref_client) + 1);

	assert(strcmp(msg_pref_client, data) == 0);
	if (test_epoll(serv)) {
		printf("this would not happen in fastopen test, but sometimes in std test, maybe the wrong implementation of epoll\n");
	}

	assert(stable_send(serv, msg_pref_server, strlen(msg_pref_server) + 1)
			== strlen(msg_pref_server) + 1);
	printf("test passed\n");
	close(serv);
	return 0;
}

static void *sthread(void *targ)
{
	int fd = (int)targ;
	return (void *)work_server(fd);
}

static int server_thread(int fd)
{
	pthread_t pid;
	if (pthread_create(&pid, NULL, sthread, (void*) fd) < 0) {
		printf("server: can not create thread for %d\n", fd);
		return -1;
	}
	if (pthread_detach(pid) < 0) {
		printf("server: error in detaching the thread for %d\n", fd);
		return -1;
	}
	return 0;
}

static void* cthread(void* targ)
{
	int fd = (int) targ;
	printf("fd is %d\n", fd);
	return (void*) work_client(fd);
}

static int client_thread(int fd)
{
	pthread_t pid;
	if (pthread_create(&pid, NULL, cthread, (void*) fd) < 0) {
		printf("client: can not create thread for %d\n", fd);
		return -1;
	}
	if (pthread_detach(pid) < 0) {
		printf("client: error in detaching the thread for %d\n", fd);
		return -1;
	}
	return 0;
}


int main(int argc, char **argv)
{
	if (argc > 1 && argv[1][0] == 'c') {
		int client_count = 1000;
		if (argc > 2) {
			client_count = atoi(argv[2]);
		}
		while (client_count-- > 0) {
			int cli = create_fastopen_client("192.168.56.101", msg_pref_client);
			client_thread(cli);
		}
		sleep (20);
	} else {
		int ser = create_fastopen_server();
		server_loop(ser, server_thread);
	}
	return 0;
}
