
#include "fastopen_kernel_def.h"
#include <signal.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "tst.h"
#include <pthread.h>

/*
 *	Thread version for the FASTOPEN connectivity test
 */


static char *msg_pref_client = "client-pref";
static char *msg_pref_server = "server-repl";

static int work_client(int fd)
{
	char buf[100];
	printf("her %d\n", fd);
	strcpy(buf, msg_pref_client);
	if (stable_send(fd, buf, strlen(buf) + 1) < 0) {
		printf("fail in first send\n");
		abort();
	}

	if (stable_recv(fd, buf, strlen(msg_pref_server) + 1) >= 0) {
		assert(strcmp(msg_pref_server, buf) == 0);
		int i = 0;
		for (i = 0; i < 1000; i++) {
			char buf_ans[100];
			sprintf(buf, "client-tst-%d", i);
			if (stable_send(fd, buf, strlen(buf) + 1) < 0)
				break;
			sprintf(buf_ans, "server-tst-%d", i);
			if (stable_recv(fd, buf, strlen(buf_ans) + 1) < 0)
				break;
			assert(strcmp(buf_ans, buf) == 0);
		}
	} else {
		printf("receive error");
	}
	close(fd);
	return 0;
}

static int work_server(int serv)
{
	unsigned char data[1024];

	if (stable_recv(serv, (char*) data, strlen(msg_pref_client) + 1) < 0) {
		printf("fail in recv the first msg\n");
		abort();
	}
	assert(strcmp(msg_pref_client, (char*) data) == 0);

	if (stable_send(serv, msg_pref_server, strlen(msg_pref_server) + 1) >=
	    0) {
		int i = 0;
		for (i = 0; i < 1000; i++) {
			char buf[100];
			char buf_ans[100];
			sprintf(buf_ans, "client-tst-%d", i);
			if (stable_recv(serv, buf, strlen(buf_ans) + 1) < 0)
				break;
			assert(strcmp(buf_ans, buf) == 0);
			sprintf(buf, "server-tst-%d", i);
			if (stable_send(serv, buf, strlen(buf) + 1) < 0)
				break;
		}
		printf("test passed\n");
	}
	close(serv);
	return 0;
}

static void *sthread(void *targ)
{
	int fd = (int)targ;
	return (void *)work_server(fd);
}

static void child_reaper(int signo)
{
	int ret;
	while ((ret = waitpid(-1, NULL, WNOHANG)) > 0)
		;
	if (ret < 0) {
		printf("error in receiving children: %s\n", strerror(errno));
	}
}

static int server_thread(int fd)
{
	pid_t pid;

	if ((pid = fork()) == 0) {
		// child
		work_server(fd);
		exit(0);
	} else if (pid > 0) {
		//parent
	} else {
		printf("fork error %s\n", strerror(errno));
		abort();
	}
	return 0;
}

static void* cthread(void* targ)
{
	int fd = (int) targ;
	printf("fd is %d\n", fd);
	return (void*) work_client(fd);
}

static int client_thread(int fd)
{
	pid_t pid;
	if ((pid = fork()) == 0) {
		// child
		work_client(fd);
		exit(0);
	} else if (pid > 0) {
		//parent
	} else {
		printf("fork error %s\n", strerror(errno));
		abort();
	}
	return 0;
}


int main(int argc, char **argv)
{
	/*
	 *	Well, have we convinced us to use the automatical block version
	 */
	signal(SIGCHLD, child_reaper);
	if (argc > 1 && argv[1][0] == 'c') {
		int client_count = 1000;
		if (argc > 2) {
			client_count = atoi(argv[2]);
		}
		while (client_count-- > 0) {
			int cli = create_std_client("192.168.56.101");
			client_thread(cli);
		}
		sleep (20);
	} else {
		int ser = create_std_server();
		server_loop(ser, server_thread);
	}
	return 0;
}
