
#include "fastopen_daemon.h"
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <sched.h>

#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <sys/time.h>
#include <stdlib.h>
#include <errno.h>
//
//	Test:
//		add proc
//		del proc
//		dup proc
//		add sock
//		del sock
//		get state
//		set state
//		cmp set state
//

#define fastopen_to_global(x) x
static double cur_time()
{
	struct timeval t;
	gettimeofday(&t, NULL);
	return t.tv_sec + t.tv_usec * 1E-6;
}

static int fastopen_cmd(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{

	int fastopen_daemon_fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (fastopen_daemon_fd < 0) {
		fprintf(stderr, "can not connect to fastopen daemon: %s\n", strerror(errno));
		exit(1);
	}

	struct sockaddr_un client_addr = {
		.sun_family = AF_UNIX
	};
	pid_t p = getpid();
	sprintf(client_addr.sun_path, "/tmp/client-%u", p);

	if (remove(client_addr.sun_path) != 0 && errno != ENOENT) {
		fprintf(stderr, "can not create client address: %s\n", strerror(errno));
		exit(1);
	}

	if (bind(fastopen_daemon_fd, (struct sockaddr*) &client_addr, sizeof(client_addr)) <0) {

		fprintf(stderr, "can not bind to fastopen daemon: %s\n", strerror(errno));
		exit(1);
	}
	struct sockaddr_un daemon_addr = {
		.sun_family = AF_UNIX
	};

	strcpy(daemon_addr.sun_path, kFASTOPEN_SERV_PATH);
	if (connect(fastopen_daemon_fd, (struct sockaddr*) &daemon_addr, sizeof(daemon_addr)) < 0) {
		fprintf(stderr, "can not connect to fastopen daemon: %s\n", strerror(errno));
		exit(1);
	}
	double start = cur_time();
	if (send(fastopen_daemon_fd, cmd, sizeof(*cmd), 0) != sizeof(*cmd)) {
		printf("stat %d\n", ret->state);
		goto err_out;
	}
	if (ret == NULL)
		goto out;
	else if (recv(fastopen_daemon_fd, ret, sizeof(*ret), 0) != sizeof(*ret)) {
		printf("stat %d\n", ret->state);
		goto err_out;
	}
	double end = cur_time();
	printf("command cost %.9lf second\n", end - start);
out:
	remove(client_addr.sun_path);
	close(fastopen_daemon_fd);
	return 0;
err_out:
	remove(client_addr.sun_path);
	close(fastopen_daemon_fd);
	return -1;
}

static int fastopen_add_proc()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_ADD_PROC,
		.pid1 = getpid()
	};
	return fastopen_cmd(&cmd, NULL);
}

static void fastopen_del_proc()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEL_PROC,
		.pid1 = getpid()
	};
	fastopen_cmd(&cmd, NULL);
}

static int fastopen_dup_proc()
{
	struct  FASTOPENCmd cmd = {
		.op = FASTOPEN_DUP_PROC,
		.pid1 = getpid(),
		.pid2 = getppid()
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_add_sock(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_ADD_SOCK,
		.gsockfd = FASTOPEN_NEW_SOCK,
		.pid1 = getpid()
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid)) {
		return -1;
	}
	return ret.state;
}

static int fastopen_del_sock(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEL_SOCK,
		.gsockfd = fastopen_to_global(fd),
		.pid1 = getpid()
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_get_sock_state(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_GET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;
	return ret.state;
}

static int fastopen_wait_connect(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_WAIT_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;		
	return ret.state;
}

static int fastopen_set_state(int fd, int state)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_SET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
		.state = state
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_cmp_set_state(int fd, int oldstate, int newstate)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_CMP_SET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
		.cmp_state = oldstate,
		.state = newstate
	};
	struct FASTOPENRet ret;
	
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid)) {
		return -1;
	}
	return ret.state;
}

static void fastopen_dump_proc_table()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEBUG_PROC_LIST
	};
	fastopen_cmd(&cmd, NULL);
}

static void fastopen_dump_sock_table()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEBUG_SOCK_LIST
	};
	fastopen_cmd(&cmd, NULL);
}

static void test_add_proc()
{
	fastopen_add_proc();
	fastopen_dump_proc_table();
	pid_t ret = fork();
	if (ret == 0) {
		fastopen_add_proc();
		fastopen_dump_proc_table();
	}
	if (ret > 0) {
		int status = 0;
		wait(&status);
	}
	fastopen_del_proc();
	fastopen_dump_proc_table();
}

static void test_add_sock()
{
	fastopen_add_proc();
	int s = fastopen_add_sock(0);
	printf("hi %d\n", s);
	fastopen_dump_proc_table();
	fastopen_dump_sock_table();
	fastopen_del_proc();
	fastopen_dump_proc_table();
	fastopen_dump_sock_table();
}


static void test_dup_proc()
{
	fastopen_add_proc();
	int s = fastopen_add_sock(0);
	printf("hi %d\n", s);
	fastopen_dump_proc_table();
	fastopen_dump_sock_table();
	pid_t ret = fork();
	if (ret == 0) {
		fastopen_add_proc();
		fastopen_dup_proc();
		fastopen_add_sock(1);
		fastopen_dump_proc_table();
		fastopen_dump_sock_table();
	}
	if (ret > 0) {
		int status = 0;
		wait(&status);
	}
	if (ret == 0) {
		fastopen_del_proc();
		fastopen_dump_proc_table();
		fastopen_dump_sock_table();
	} else {
		fastopen_del_sock(s);
		fastopen_dump_proc_table();
		fastopen_dump_sock_table();
		fastopen_del_proc();
	}
}

static void test_cmp_set_state()
{
	fastopen_add_proc();
	int s = fastopen_add_sock(0);

	int ret = fork();
	if (ret == 0) {
		fastopen_add_proc();
		fastopen_dup_proc();

		fastopen_dump_sock_table();
		sleep(1);

		int old = fastopen_cmp_set_state(s, FASTOPEN_SOCK_IDLE, FASTOPEN_SOCK_CONNECTING);
		if (old == FASTOPEN_SOCK_IDLE) {
			fastopen_set_state(s, FASTOPEN_SOCK_INVALID);
			printf("done invalid!\n");
		} else {
			int state = fastopen_get_sock_state(s);
			printf("before wait state: %d\n", state);
			state = fastopen_wait_connect(s);
			printf("after wait state: %d\n", state);
		}
		fastopen_dump_sock_table();
	} else {
		int old = fastopen_cmp_set_state(s, FASTOPEN_SOCK_IDLE, FASTOPEN_SOCK_CONNECTING);
		if (old == FASTOPEN_SOCK_IDLE) {
			fastopen_set_state(s, FASTOPEN_SOCK_CONNECTED);
			printf("done connected!\n");
		} else {
			int state = fastopen_get_sock_state(s);
			printf("before wait state: %d\n", state);
			state = fastopen_wait_connect(s);
			printf("after wait state: %d\n", state);
		}
		fastopen_dump_sock_table();
		wait(NULL);
	}
	fastopen_del_proc();
}


int main()
{
//	fastopen_dump_sock_table();
//	test_add_proc();
//	test_dup_proc();
//	test_add_sock();
//	test_cmp_set_state();
	fastopen_dump_sock_table();
	fastopen_dump_proc_table();
	return 0;
}

